from bewe.alpha_seeker import base_data
from bewe.services.databases import asset_orm
from bewe.services.databases import db_manager
import pandas as pd
import numpy as np
import tensorflow as tf


class DailyStock(base_data.DataContainer):

    def __init__(self):
        super().__init__(asset_name='daily_stock')
        self.db = db_manager.DBManager()
        self.format = base_data.DataFormat(
            required_field=['feature', 'label']
        )

    def query_stock(self, stock_id, limit=100):
        records = self.db.query(
            asset_orm.StockPrice,
            conditions={'stock_id': stock_id},
            order_conditions={'date': False},
            limit=limit
        )
        records = [r.get_dict() for r in records]
        df = pd.DataFrame(records).sort_values('date')
        return df

    def get_level(self, value):
        return min(int(value*100) + 10, 10)

    def transform(self, data: pd.DataFrame):
        self.data = []

        data['actionable_price'] = data['open_price'].shift(-1)
        data['val_start'] = data['close_price'].shift(-1)
        data['val_end'] = data['close_price'].shift(-6)
        data = data[~data['val_end'].isna()]
        data['daily_diff'] = data.apply(
            lambda row: self.get_level(
                (row.close_price-row.open_price)/row.open_price
            ),
            axis=1
        )
        data['label'] = data.apply(
            lambda row: self.get_level(
                (row.val_end-row.val_start)/row.val_start
            ),
            axis=1
        )

        data['feat'] = [window.to_list() for window in data.daily_diff.rolling(window=10)]
        data['quality'] = [len(r) >= 10 for r in data.feat.values]
        data = data[data['quality']]

        for _, row in data.iterrows():
            row = base_data.DataUnit(
                format=self.format,
                data={
                    'feature': row.feat,
                    'label': row.label
                },
                actionable_price=row['actionable_price']
            )
            self.data.append(row)
        self.size = len(self.data)


if __name__ == '__main__':

    features = []
    labels = []

    for stock in ['2330', '4919', '0050', '0056', '2356', '3231']:
        loader = DailyStock()
        loader_data = loader.query_stock(stock, limit=200)
        loader.transform(loader_data)

        for row in loader:
            features.append(row.data['feature'])
            labels.append(row.data['label'])

    features = np.array(features)
    labels = np.array(labels)
    shuffle_index = np.arange(len(labels))
    np.random.shuffle(shuffle_index)

    features = features[shuffle_index]
    labels = labels[shuffle_index]

    import collections
    print(collections.Counter(labels))

    tf.keras.backend.clear_session()
    model = tf.keras.models.Sequential()
    model.add(tf.keras.layers.Embedding(20, 32))
    model.add(tf.keras.layers.LSTM(32))
    model.add(tf.keras.layers.Dense(16, activation='relu'))
    model.add(tf.keras.layers.Dense(16, activation='relu'))
    model.add(tf.keras.layers.Dense(20, activation='softmax'))
    model.compile(loss='sparse_categorical_crossentropy', optimizer='adam', metrics=['acc'])

    model.fit(features, labels, epochs=10, validation_split=0.1)



import datetime
import threading

from absl import logging
from bewe.services.crawler import secret_manager
from bewe.services.databases import asset_orm
from bewe.services.databases import db_manager
from FinMind.data import data_loader
from typing import Optional

_FINMIND_KEY = 'FINMIND_API'


class FinmindWrapper:
    def __init__(self):
        self.token = secret_manager.get_secret_tokens(_FINMIND_KEY)
        self.data_loader = data_loader.DataLoader()
        self.data_loader.login_by_token(self.token)
        self.db_manager = db_manager.DBManager()

    def get_tw_stock_info(self):
        df = self.data_loader.taiwan_stock_info()

        stocks, categories = [], []
        for _, row in df.iterrows():
            stock = asset_orm.Stock(
                stock_id=row.stock_id,
                stock_name=row.stock_name
            )

            category = asset_orm.Category(
                stock_id=row.stock_id,
                category=row.industry_category
            )

            stocks.append(stock)
            categories.append(category)

        self.db_manager.update(stocks)
        self.db_manager.update(categories)

    def incremental_get_stock_price(self, limit: int = 2):
        stocks = self.db_manager.query(
            asset_orm.Stock,
            conditions={'flag': True},
            order_conditions={'last_execution': True},
            limit=limit)

        jobs = []

        for stock in stocks:
            job = threading.Thread(target=self.get_stock_price, args=(stock.stock_id, stock.last_price_date))
            job.start()
            jobs.append(job)

        for job in jobs:
            job.join()
        return

    def get_stock_price(
            self,
            stock_id: str,
            start_date: datetime.datetime,
            end_date: Optional[datetime.datetime] = None):
        if not end_date:
            end_date = datetime.datetime.now()

        date_diff = (end_date - start_date).days
        start_date = start_date.strftime('%Y-%m-%d')
        end_date = end_date.strftime('%Y-%m-%d')

        if date_diff == 0:
            logging.info('Start and End are on the same date. Skip the query.')
            return

        logging.info(f'[Execution] get stock price: {stock_id}, {start_date}~{end_date}.')

        try:
            df = self.data_loader.taiwan_stock_daily(
                stock_id=stock_id,
                start_date=start_date,
                end_date=end_date
            )

            prices = []
            stock = self.db_manager.query(asset_orm.Stock, {'stock_id': stock_id})[0]
            stock.last_execution = datetime.datetime.now()

            for _, row in df.iterrows():

                price = asset_orm.StockPrice(
                    stock_id=row.stock_id,
                    date=row.date,
                    open_price=row.open,
                    close_price=row.close,
                    max_price=row['max'],
                    min_price=row['min'],
                    volume=row.Trading_Volume
                )
                prices.append(price)

            if prices:
                max_price_date = datetime.datetime.strptime(
                    max(df.date.values), '%Y-%m-%d')
                stock.last_price_date = max_price_date
                self.db_manager.update(prices, flush=False)
            if not prices and date_diff >= 10:
                stock.flag = False
            self.db_manager.update([stock])

        except Exception:
            logging.exception('API reach limitation.')
